"""
Helpers and extensions for sphinx.

This subpackage contains two sphinx directives:

* ``.. generate::`` which includes raw output generated by a Python script into
  a ``.. raw::`` block. This can be used to for example, include HTML output
  into the built docs.

* ``.. changelog::`` which renders the latest changelog with towncrier and
  includes it in the documentation.

To use them add ``'sunpy.util.sphinx.generate'`` and
``'sunpy.util.sphinx.changelog'`` to the ``extensions`` list in your ``conf.py``
file.

"""
