# contains links to HELIO registries
"""
The following link is the link to the registry, should it be necessary, this
link can easily be replaced or swapped.
"""
LINK = 'http://helio.ukssdc.ac.uk/helio_registry/viewResourceEntry_body.jsp?XML=true&IVORN=ivo://helio-vo.eu'
